Unit XMLDataset;

{
Version: 1.001
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.

The original code is XMLDataset.pas, released March 14, 2001.

The developer of this code is Basri Kanca (basrikanca@excite.com)
(C) 2001 Basri Kanca. All Rights Reserved.

See the accompanying XMLDataset.txt for further info.

}

Interface

Uses
  Windows,
  SysUtils,
  Messages,
  Classes,
  Db,
  rjMime, // RJ
  XMLDatasetUtils;

{.$DEFINE TTABLE_DERIVATIVE}

Type
  EXMLDatasetError = Exception;

  TXMLDataset = Class(TComponent)
  Private
    // RJ FBase64MIME: TBase64MIME;
    FDataSet: TDataSet;
    FDummyBoolean: Boolean;
    FDummyDate: TDateTime;
    FDummyInteger: Integer;
    FDummyName: String;
    FDummySingle: Single;
    FFieldDefCount: Integer;
    FFileName: String;
    FFileVersion: Single;
    FHasChanges: Boolean;
    FHasFieldDefs: Boolean;
    FHasIndexDefs: Boolean;
    FHasRecordData: Boolean;
    FIndexDefCount: Integer;
    FMemoryStream: TMemoryStream;
    FPacketTime: TDateTime;
    FPacketVersion: Single;
    FRecordCount: Integer;
    FSeekInXMLFile: TSeekInXMLFile;
    FTableName: String;

    Function Execute: Boolean;
    Function GetBounds(Const ABounds: Array Of TBoundsInfo; Const ABeginPos, AEndPos: Integer; Var ABoundsBegin, ABoundsEnd: Integer): Boolean;
    Function GetText: String;
    Procedure GetDataPacketProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
    Procedure GetMetaDataProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
    Procedure GetFieldDefinitionProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
    Procedure GetRecordDataProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
    Procedure GetIndexDefinitionProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
    Procedure DoAtribAndValue(Var ABeginPos: Integer; AEndPos: Integer; Var AAttribCount: Integer; Var AArray: Array Of TParamAndValue);
    Procedure DoFieldDefinitions(Var ABeginPos: Integer; Const AEndPos: Integer);
    Procedure DoIndexDefinitions(Var ABeginPos: Integer; Const AEndPos: Integer); //I have not tested this routine
    Procedure DoRecordData(Var ABeginPos: Integer; Const AEndPos: Integer);
    Procedure DoFieldData(Var ABeginPos: Integer; Const AEndPos: Integer);
    Procedure FindFirst(Const ASubstring: String; Const ABeginPos, AEndPos: Integer; Var AResult: Integer);
    Procedure SetDataset(ADataSet: TDataSet);
    Procedure SetText(AValue: String);
    Procedure CopyToString(ABeginPos, AEndPos: Integer; Var AResult: String);
  Protected
    Procedure Notification(AComponent: TComponent; Operation: TOperation); Override; // Gmarra <m_gamarra@bol.com.br>
  Public
    Constructor Create(AOwner: TComponent); Override;
    Destructor Destroy; Override;
    Procedure Clear;
    Procedure LoadFromFile(AFileName: String);
    Procedure LoadFromStream(AStream: TStream);
    Procedure SaveToFile(AFileName: String);
    Procedure SaveToStream(AStream: TStream);
    Property Text: String Read GetText Write SetText;
  Published
    Property TableName: String Read FTableName Write FDummyName;
    Property DataSet: TDataSet Read FDataSet Write SetDataset;
    Property FileVersion: Single Read FFileVersion Write FDummySingle;
    Property PacketTime: TDateTime Read FPacketTime Write FDummyDate;
    Property PacketVersion: Single Read FPacketVersion Write FPacketVersion;
    Property RecordCount: Integer Read FRecordCount Write FDummyInteger;
    Property FieldDefCount: Integer Read FFieldDefCount Write FDummyInteger;
    Property IndexDefCount: Integer Read FIndexDefCount Write FDummyInteger;
    Property HasFieldDefs: Boolean Read FHasFieldDefs Write FDummyBoolean;
    Property HasIndexDefs: Boolean Read FHasIndexDefs Write FDummyBoolean;
    Property HasRecordData: Boolean Read FHasRecordData Write FDummyBoolean;
    Property HasChanges: Boolean Read FHasChanges Write FDummyBoolean;
  End;

Procedure Register;

Implementation

Procedure Register;
Begin
  RegisterComponents('XMLDataset', [TXMLDataset]);
End;

Constructor TXMLDataset.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  // RJ FBase64MIME := TBase64MIME.Create (nil);
  FMemoryStream := TMemoryStream.Create;
  FSeekInXMLFile := TSeekInXMLFile.Create(Nil);
  FDataSet := Nil;
  FPacketTime := 0.0;
  FPacketVersion := 0.0;
  FFileVersion := FILE_VERSION;
End;

Destructor TXMLDataset.Destroy;
Begin
  FSeekInXMLFile.Free;
  // RJ FBase64MIME.Free;
  FMemoryStream.Free;
  FDataSet := Nil;
  Inherited Destroy;
End;

Procedure TXMLDataset.Notification(AComponent: TComponent; Operation: TOperation);
Begin
  Inherited Notification(AComponent, Operation);
  If Operation <> opRemove Then Exit;
  If AComponent = FDataSet Then FDataSet := Nil;
End;

Procedure TXMLDataset.Clear;
Begin
  FSeekInXMLFile.Clear;
  FMemoryStream.Clear;
End;

Function TXMLDataset.GetText: String;
Begin
  Result := FSeekInXMLFile.Text;
End;

Procedure TXMLDataset.SetText(AValue: String);
Var
  Size1: Integer;
Begin
  FFileName := '';
  Clear;
  Size1 := Length(AValue);
  FMemoryStream.Write(PChar(AValue)^, Size1);
  FSeekInXMLFile.Text := PChar(FMemoryStream.Memory);
  Execute;
  FSeekInXMLFile.Clear;
End;

Procedure TXMLDataset.LoadFromStream(AStream: TStream);
Begin
  If Not Assigned(FDataSet) Then Raise EXMLDatasetError.Create('Dataset not assigned');
  If Not Assigned(AStream) Then Raise EXMLDatasetError.Create('Dataset not assigned');
  Clear;
  FMemoryStream.CopyFrom(AStream, 0);
  FSeekInXMLFile.Text := PChar(FMemoryStream.Memory);
  Execute;
  FSeekInXMLFile.Clear;
End;

Procedure TXMLDataset.LoadFromFile(AFileName: String);
Begin
  If Not Assigned(FDataSet) Then Raise EXMLDatasetError.Create('Dataset not assigned');
  If Not FileExists(AFileName) Then Raise EXMLDatasetError.Create('File does not exist');
  Clear;
  FSeekInXMLFile.FileName := AFileName;
  FFileName := AFileName;
  Execute;
  FSeekInXMLFile.Clear;
End;

Procedure TXMLDataset.SaveToStream(AStream: TStream);
  Procedure StringToStream(Const AString: String);
  Var
    String1: String;
    length1: Integer;
  Begin
    String1 := AString + CRLF_CHARS;
    length1 := Length(String1);
    AStream.Write(PChar(String1)^, length1);
  End;
Var
  Index1: Integer;
  Index2: Integer;
  Size1: Integer;
  FieldCount1: Integer;
  IndexCount1: Integer;
  RecordCount1: Integer;
  Filtered1: Boolean;
  Active1: Boolean;
  ControlsDisabled1: Boolean;
  Hour1: Word;
  Min1: Word;
  Sec1: Word;
  MSec1: Word;
  Day1: Word;
  Month1: Word;
  Year1: Word;
  Bookmark1: TBookmark;

  String1: String;
  String2: String;
  FieldLine1: String;
  Field1: TField;
{$IFDEF TTABLE_DERIVATIVE}
  IndexDef1 TIndexDef;
{$ENDIF}
  DecimalSeparator1: Char;
  ShortDateFormat1: String;
  ShortTimeFormat1: String;
Begin
  If Not Assigned(FDataSet) Then Raise EXMLDatasetError.Create('Dataset not assigned');
  If Not Assigned(AStream) Then Raise EXMLDatasetError.Create('Dataset not assigned');
  DecimalSeparator1 := DecimalSeparator;
  ShortDateFormat1 := ShortDateFormat;
  ShortTimeFormat1 := ShortTimeFormat;
  DecodeDate(Now, Year1, Month1, Day1);
  DecodeTime(Now, Hour1, Min1, Sec1, MSec1);
  Try
    DecimalSeparator := '.';
    ShortDateFormat := 'yyyy-mm-dd';
    ShortTimeFormat := 'hh:nn:ss:zzz';
    Active1 := FDataSet.Active;
    If Not Active1 Then FDataSet.Open;
    Bookmark1 := FDataSet.GetBookmark;
    ControlsDisabled1 := FDataSet.ControlsDisabled;
    Filtered1 := FDataSet.Filtered;
    If Not ControlsDisabled1 Then FDataSet.DisableControls;
    RecordCount1 := FDataSet.RecordCount;
    FieldCount1 := FDataSet.FieldCount;
{$IFDEF TTABLE_DERIVATIVE}
    IndexCount1 := FDataSet.IndexDefs.Count;
    FTableName := FDataSet.TableName;
{$ELSE}
    IndexCount1 := 0;
{$ENDIF}
    StringToStream(XML_BEGIN + SPACE_CHAR +
      VERSION_LABEL + EQUAL_CHAR + QUOTE_CHAR + '1.0' + QUOTE_CHAR + SPACE_CHAR +
      STANDALONE_LABEL + EQUAL_CHAR + QUOTE_CHAR + YES_LABEL + QUOTE_CHAR + XML_END);

    StringToStream(
      DATAPACKET_BEGIN + SPACE_CHAR +
      VERSION_LABEL + EQUAL_CHAR + QUOTE_CHAR + Format('%3.3f', [FFileVersion]) + QUOTE_CHAR + SPACE_CHAR +
      DESTINATION_LABEL + EQUAL_CHAR + QUOTE_CHAR + FTableName + QUOTE_CHAR + SPACE_CHAR +
      DAY_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Day1) + QUOTE_CHAR + SPACE_CHAR +
      MONTH_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Month1) + QUOTE_CHAR + SPACE_CHAR +
      YEAR_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Year1) + QUOTE_CHAR + SPACE_CHAR +
      HOUR_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Hour1) + QUOTE_CHAR + SPACE_CHAR +
      MIN_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Min1) + QUOTE_CHAR + SPACE_CHAR +
      SEC_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Sec1) + QUOTE_CHAR + SPACE_CHAR +
      MSEC_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(MSec1) + QUOTE_CHAR + TAG_END);

    If IndexCount1 > 0 Then Begin
      StringToStream(METADATA_BEGIN + SPACE_CHAR +
        FIELDDEFS_LABEL + EQUAL_CHAR + QUOTE_CHAR + YES_LABEL + QUOTE_CHAR + SPACE_CHAR +
        INDEXDEFS_LABEL + EQUAL_CHAR + QUOTE_CHAR + YES_LABEL + QUOTE_CHAR + SPACE_CHAR +
        RECORDDATA_LABEL + EQUAL_CHAR + QUOTE_CHAR + YES_LABEL + QUOTE_CHAR + SPACE_CHAR +
        CHANGES_LABEL + EQUAL_CHAR + QUOTE_CHAR + YES_LABEL + QUOTE_CHAR + TAG_END);
    End Else Begin
      StringToStream(METADATA_BEGIN + SPACE_CHAR +
        FIELDDEFS_LABEL + EQUAL_CHAR + QUOTE_CHAR + YES_LABEL + QUOTE_CHAR + SPACE_CHAR +
        INDEXDEFS_LABEL + EQUAL_CHAR + QUOTE_CHAR + NO_LABEL + QUOTE_CHAR + SPACE_CHAR +
        RECORDDATA_LABEL + EQUAL_CHAR + QUOTE_CHAR + YES_LABEL + QUOTE_CHAR + SPACE_CHAR +
        CHANGES_LABEL + EQUAL_CHAR + QUOTE_CHAR + YES_LABEL + QUOTE_CHAR + TAG_END);
    End;

    Try // Setup flags for fields to save.
      StringToStream(FIELDDEFS_BEGIN + SPACE_CHAR + COUNT_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(FieldCount1) + QUOTE_CHAR + TAG_END);

      For Index1 := 0 To FieldCount1 - 1 Do Begin
        Field1 := FDataSet.Fields[Index1];
        String1 := (FIELDDEF_BEGIN + SPACE_CHAR + NAME_LABEL + EQUAL_CHAR + QUOTE_CHAR + Field1.FieldName + QUOTE_CHAR);
        String1 := String1 + SPACE_CHAR + FIELDKIND_LABEL + EQUAL_CHAR + QUOTE_CHAR;
        Case Field1.FieldKind Of
          fkData: String1 := String1 + 'data';
          fkCalculated: String1 := String1 + 'calculated';
          fkLookup: String1 := String1 + 'lookup';
          fkInternalCalc: String1 := String1 + 'internalcalc';
          fkAggregate: String1 := String1 + 'aggregate';
          Else
            String1 := String1 + 'unknown';
        End;
        String1 := String1 + QUOTE_CHAR;
        String1 := String1 + (SPACE_CHAR + DATATYPE_LABEL + EQUAL_CHAR + QUOTE_CHAR + DataTypeToString(FDataSet.Fields[Index1]) + QUOTE_CHAR);
        String1 := String1 + (SPACE_CHAR + FIELDSIZE_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Field1.Size) + QUOTE_CHAR);
        String1 := String1 + (SPACE_CHAR + DISPLAYLABEL_LABEL + EQUAL_CHAR + QUOTE_CHAR + Field1.DisplayLabel + QUOTE_CHAR);
        String1 := String1 + (SPACE_CHAR + EDITMASK_LABEL + EQUAL_CHAR + QUOTE_CHAR + Field1.EditMask + QUOTE_CHAR);
        String1 := String1 + (SPACE_CHAR + DISPLAYWIDTH_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Field1.DisplayWidth) + QUOTE_CHAR);
        String1 := String1 + (SPACE_CHAR + FIELDINDEX_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Field1.Index) + QUOTE_CHAR);
        If Field1.Required Then String1 := String1 + (SPACE_CHAR + REQUIRED_LABEL + EQUAL_CHAR + QUOTE_CHAR + TRUE_LABEL + QUOTE_CHAR) Else String1 := String1 + (SPACE_CHAR + REQUIRED_LABEL + EQUAL_CHAR + QUOTE_CHAR + FALSE_LABEL + QUOTE_CHAR);
        If Field1.ReadOnly Then String1 := String1 + (SPACE_CHAR + READONLY_LABEL + EQUAL_CHAR + QUOTE_CHAR + TRUE_LABEL + QUOTE_CHAR) Else String1 := String1 + (SPACE_CHAR + READONLY_LABEL + EQUAL_CHAR + QUOTE_CHAR + FALSE_LABEL + QUOTE_CHAR);
        StringToStream(String1 + FIELDDEF_END);
      End;
      StringToStream(FIELDDEFS_END);

      StringToStream(INDEXDEFS_BEGIN + SPACE_CHAR + COUNT_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(IndexCount1) + QUOTE_CHAR + TAG_END);
{$IFDEF TTABLE_DERIVATIVE}
      For Index1 := 0 To IndexCount1 - 1 Do Begin
        IndexDef1 := FDataSet.IndexDefs.Items[Index1];
        String1 := INDEX_BEGIN + SPACE_CHAR + NAME_LABEL + EQUAL_CHAR + QUOTE_CHAR + IndexDef1.Name + QUOTE_CHAR + SPACE_CHAR;
        String1 := String1 + INDEXFIELDS_LABEL + EQUAL_CHAR + QUOTE_CHAR + IndexDef1.Fields + QUOTE_CHAR + SPACE_CHAR;
        String1 := String1 + DISPLAYNAME_LABEL + EQUAL_CHAR + QUOTE_CHAR + IndexDef1.DisplayLabel + QUOTE_CHAR + SPACE_CHAR;
        String1 := String1 + INDEXNO_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Index1) + QUOTE_CHAR + SPACE_CHAR;
        If ixDescending In IndexDef1.Options Then String1 := String1 + SORTORDER_LABEL + EQUAL_CHAR + QUOTE_CHAR + DESCENDING_LABEL + QUOTE_CHAR + SPACE_CHAR Else String1 := String1 + SORTORDER_LABEL + EQUAL_CHAR + QUOTE_CHAR + ASCENDING_LABEL + QUOTE_CHAR + SPACE_CHAR;
        If ixCaseInsensitive In IndexDef1.Options Then String1 := String1 + CASESENSITIVE_LABEL + EQUAL_CHAR + QUOTE_CHAR + TRUE_LABEL + QUOTE_CHAR Else String1 := String1 + CASESENSITIVE_LABEL + EQUAL_CHAR + QUOTE_CHAR + FALSE_LABEL + QUOTE_CHAR;
        If ixNonMaintained In IndexDef1.Options Then String1 := String1 + NONMAINTAINED_LABEL + EQUAL_CHAR + QUOTE_CHAR + TRUE_LABEL + QUOTE_CHAR Else String1 := String1 + NONMAINTAINED_LABEL + EQUAL_CHAR + QUOTE_CHAR + FALSE_LABEL + QUOTE_CHAR;
        If ixUnique In IndexDef1.Options Then String1 := String1 + UNIQUE_LABEL + EQUAL_CHAR + QUOTE_CHAR + TRUE_LABEL + QUOTE_CHAR Else String1 := String1 + UNIQUE_LABEL + EQUAL_CHAR + QUOTE_CHAR + FALSE_LABEL + QUOTE_CHAR;
        StringToStream(String1 + EXTRATAG_END);
      End;
{$ENDIF}
      StringToStream(INDEXDEFS_END);
      StringToStream(METADATA_END);

      StringToStream(RECORDDATA_BEGIN + SPACE_CHAR + COUNT_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(RecordCount1) + QUOTE_CHAR + TAG_END);
      FDataSet.First;
      While Not FDataSet.EoF Do Begin
        StringToStream(ROW_BEGIN);
        For Index2 := 0 To FieldCount1 - 1 Do Begin
          Field1 := FDataSet.Fields[Index2];
          FieldLine1 := FIELD_LABEL + SPACE_CHAR + NAME_LABEL + EQUAL_CHAR + QUOTE_CHAR + Field1.FieldName + QUOTE_CHAR;
          If Field1.IsNull Then StringToStream(TAG_BEGIN + FieldLine1 + SPACE_CHAR + SIZE_LABEL + EQUAL_CHAR + QUOTE_CHAR + '-1' + QUOTE_CHAR + EXTRATAG_END)
          Else If (Field1 Is TObjectField) Then Raise EXMLDatasetError.Create('ObjectFields are not handled!')
{$IFDEF ver130}
          Else If (Field1 Is TInterFaceField) Then Raise EXMLDatasetError.Create('InterFaceFields are not handled!')
{$ENDIF}
          Else If (Field1 Is TBinaryField) Then Raise EXMLDatasetError.Create('BinaryFields are not handled!')
          Else If (Field1 Is TNumericField) Or (Field1 Is TBooleanField) Then Begin
            String1 := Field1.AsString;
            StringToStream(TAG_BEGIN + FieldLine1 + SPACE_CHAR + VALUE_LABEL + EQUAL_CHAR + QUOTE_CHAR + String1 + QUOTE_CHAR + FORWARD_SLASH + TAG_END);
          End Else Begin
            String1 := Field1.AsString;
            Size1 := Length(String1);
            StringToStream(TAG_BEGIN + FieldLine1 + SPACE_CHAR + SIZE_LABEL + EQUAL_CHAR + QUOTE_CHAR + IntToStr(Size1) + QUOTE_CHAR + TAG_END);
            If Size1 > 0 Then Begin
              // RJ string2 := '';
              // RJ FBase64MIME.Encode (string1, string2);
              String2 := MimeEncodeString(String1);
              StringToStream(CDATA_BEGIN);
              StringToStream(String2);
              StringToStream(CDATA_END);
            End;
            StringToStream(FIELD_END_FULL);
          End;
        End;
        StringToStream(ROW_END);
        FDataSet.Next;
      End;
      StringToStream(RECORDDATA_END);
      FDataSet.GotoBookmark(Bookmark1);
      FDataSet.Active := Active1;
      StringToStream(DATAPACKET_END);
      FDataSet.Filtered := Filtered1;
    Except
      On e: Exception Do ;
    End;
    If Not ControlsDisabled1 Then FDataSet.EnableControls;
  Finally
    DecimalSeparator := DecimalSeparator1;
    ShortDateFormat := ShortDateFormat1;
    ShortTimeFormat := ShortTimeFormat1;
  End;
End;

Procedure TXMLDataset.SaveToFile(AFileName: String);
Var
  Stream1: TStream;
Begin
  If Not Assigned(FDataSet) Then Raise EXMLDatasetError.Create('Dataset not assigned');
  Stream1 := TFileStream.Create(AFileName, fmCreate Or fmShareExclusive);
  FFileName := AFileName;
  Try
    SaveToStream(Stream1);
  Finally
    Stream1.Free;
  End;
End;

Procedure TXMLDataset.SetDataset(ADataSet: TDataSet);
Begin
  If FDataSet = ADataSet Then Exit;
  FDataSet := ADataSet;
End;

Procedure TXMLDataset.FindFirst(Const ASubstring: String; Const ABeginPos, AEndPos: Integer; Var AResult: Integer);
Begin
  Try
    FSeekInXMLFile.FindFirst(ASubstring, ABeginPos, AEndPos, AResult);
  Except
  End;
End;

Procedure TXMLDataset.CopyToString(ABeginPos, AEndPos: Integer; Var AResult: String);
Begin
  FSeekInXMLFile.CopyToString(ABeginPos + 1, AEndPos - 1, AResult);
End;

Type
  TLocalFieldList = Record
    AttribCount: Integer;
    Contents: Array[0..255] Of TParamAndValue;
  End;

Procedure TXMLDataset.DoAtribAndValue(Var ABeginPos: Integer; AEndPos: Integer; Var AAttribCount: Integer; Var AArray: Array Of TParamAndValue);
  Procedure LocalCopyToString(AString: String; ABeginPos, AEndPos: Integer; Var AResult: String);
  Var
    i: Integer;
  Begin
    AResult := '';
    For i := ABeginPos + 1 To AEndPos - 1 Do AResult := AResult + AString[i];
  End;

  Procedure LocalFindFirst(Const AChar: Char; Const AString: String; Const ABeginPos, AEndPos: Integer; Var AResult: Integer);
  Var
    tI1: Integer;
  Begin
    AResult := -1;
    For tI1 := ABeginPos To AEndPos Do Begin
      If AString[tI1] = AChar Then Begin
        AResult := tI1;
        Exit;
      End;
    End;
  End;

  Procedure LocalDoAtribAndValue(Const AString: String; Var AAttribCount: Integer; Var AArray: Array Of TParamAndValue);
  Var
    EndPos1: Integer;
    Equal1: Integer;
    Space1: Integer;
    Quote1: Integer;
    Quote2: Integer;
    AttribName1: String;
    AttribValue1: String;
  Begin
    EndPos1 := Length(AString);
    LocalFindFirst(SPACE_CHAR, AString, 1, EndPos1, Space1);
    While (Space1 > 0) Do Begin
      LocalFindFirst(EQUAL_CHAR, AString, Space1, EndPos1, Equal1);
      If Equal1 > 0 Then Begin // find the next EQUAL_CHAR
        LocalCopyToString(AString, Space1, Equal1, AttribName1);
        LocalFindFirst(QUOTE_CHAR, AString, Equal1, EndPos1, Quote1);
        If Quote1 > 0 Then Begin // find beginnig of attribute value
          LocalFindFirst(QUOTE_CHAR, AString, Quote1 + 1, EndPos1, Quote2);
          If Quote2 > 0 Then Begin // find end of attribute value
            LocalCopyToString(AString, Quote1, Quote2, AttribValue1);
            AArray[AAttribCount].Name := AttribName1;
            AArray[AAttribCount].Value := AttribValue1;
            Inc(AAttribCount);
            LocalFindFirst(SPACE_CHAR, AString, Quote2, EndPos1, Space1);
          End Else Break;
        End Else Break;
      End Else Break;
    End;
  End;
Var
  String1: String;
Begin //Beginning and ending tags are NOT expected in here
  AAttribCount := 0;
  If (ABeginPos >= AEndPos) Then Exit;
  If (AEndPos > ABeginPos) And (ABeginPos < AEndPos) Then Begin
    CopyToString(ABeginPos, AEndPos, String1);
    LocalDoAtribAndValue(String1, AAttribCount, AArray);
  End;
  If String1[Length(String1)] = '/' Then ABeginPos := AEndPos + 1
  Else ABeginPos := AEndPos;
End;

Procedure TXMLDataset.DoFieldDefinitions(Var ABeginPos: Integer; Const AEndPos: Integer);
  Procedure DestroyFields;
  Begin
    FDataSet.Fields.Clear;
    If Assigned(FDataSet.DataSetField) Then FDataSet.DataSetField.Free;
  End;

  Procedure CreateFields;
  Var
    i: Integer;
  Begin
    If FDataSet.ObjectView Then Begin
      For i := 0 To FDataSet.FieldDefs.Count - 1 Do Begin
        If (FDataSet.FieldDefs[i].DataType <> ftUnknown) Then Begin
          If Not (faHiddenCol In FDataSet.FieldDefs[i].Attributes) Then Begin
            If Not FDataSet.FieldDefs.HiddenFields Then FDataSet.FieldDefs[i].CreateField(Self);
          End;
        End;
      End;
    End Else Begin
      For i := 0 To FDataSet.FieldDefList.Count - 1 Do Begin
        If (FDataSet.FieldDefList[i].DataType <> ftUnknown) Then Begin
          If Not (FDataSet.FieldDefList[i].DataType In ObjectFieldTypes) Then Begin
            If Not (faHiddenCol In FDataSet.FieldDefList[i].Attributes) Then Begin
              If Not FDataSet.FieldDefs.HiddenFields Then Begin
                FDataSet.FieldDefList[i].CreateField(Self, Nil, FDataSet.FieldDefList.Strings[i]);
              End;
            End;
          End;
        End;
      End;
    End;
  End;
Var
  FieldData1: Array[0..255] Of TLocalFieldList;
  FieldCount1: Integer;
  tI1: Integer;
  tI2: Integer;
  BeginPos1: Integer;
  EndPos1: Integer;
  FieldName1: String;
  FieldKind1: String;
  DataType1: String;
  FieldSize1: String;
  DisplayLabel1: String;
  EditMask1: String;
  DisplayWidth1: String;
  Required1: String;
  ReadOnly1: String;
  FieldKind2: TFieldKind;
  DataType2: TFieldType;
  FieldSize2: Integer;
  DisplayLabel2: String;
  EditMask2: String;
  DisplayWidth2: Integer;
  FieldNo1: String;
  Required2: Boolean;
  ReadOnly2: Boolean;
  Field1: TField;

Begin
  FDataSet.Close;
  FDataSet.FieldDefs.Clear;
  DestroyFields;

  FieldCount1 := -1;
  BeginPos1 := ABeginPos;
  While (BeginPos1 < AEndPos) Do Begin
    If GetBounds(FIELDDEF_BOUNDS, ABeginPos, AEndPos, BeginPos1, EndPos1) Then Begin
      Inc(FieldCount1);
      DoAtribAndValue(BeginPos1, EndPos1, FieldData1[FieldCount1].AttribCount, FieldData1[FieldCount1].Contents);
    End Else Break;
    ABeginPos := BeginPos1;
  End;

  For tI1 := 0 To FieldCount1 Do Begin
    For tI2 := 0 To FieldData1[tI1].AttribCount - 1 Do Begin
      If FieldData1[tI1].Contents[tI2].Name = NAME_LABEL Then FieldName1 := FieldData1[tI1].Contents[tI2].Value
      Else If FieldData1[tI1].Contents[tI2].Name = DATATYPE_LABEL Then DataType1 := FieldData1[tI1].Contents[tI2].Value
      Else If FieldData1[tI1].Contents[tI2].Name = FIELDSIZE_LABEL Then FieldSize1 := FieldData1[tI1].Contents[tI2].Value
      Else If FieldData1[tI1].Contents[tI2].Name = REQUIRED_LABEL Then Required1 := FieldData1[tI1].Contents[tI2].Value;
    End;
    DataType2 := StringToDataType(DataType1);
    If FieldSize1 = '' Then FieldSize2 := 0 Else FieldSize2 := StrToInt(FieldSize1);
    If Required1 = TRUE_LABEL Then Required2 := True Else Required2 := False;
    FDataSet.FieldDefs.Add(FieldName1, DataType2, FieldSize2, Required2);
  End;
  CreateFields;

  For tI1 := 0 To FieldCount1 Do Begin
    FieldName1 := '';
    FieldKind1 := '';
    DisplayLabel1 := '';
    EditMask1 := '';
    DisplayWidth1 := '';
    FieldNo1 := '';
    ReadOnly1 := '';

    For tI2 := 0 To FieldData1[tI1].AttribCount - 1 Do Begin
      If FieldData1[tI1].Contents[tI2].Name = NAME_LABEL Then FieldName1 := FieldData1[tI1].Contents[tI2].Value;
      If FieldData1[tI1].Contents[tI2].Name = FIELDKIND_LABEL Then FieldKind1 := FieldData1[tI1].Contents[tI2].Value;
      If FieldData1[tI1].Contents[tI2].Name = DISPLAYLABEL_LABEL Then DisplayLabel1 := FieldData1[tI1].Contents[tI2].Value;
      If FieldData1[tI1].Contents[tI2].Name = EDITMASK_LABEL Then EditMask1 := FieldData1[tI1].Contents[tI2].Value;
      If FieldData1[tI1].Contents[tI2].Name = DISPLAYWIDTH_LABEL Then DisplayWidth1 := FieldData1[tI1].Contents[tI2].Value;
      If FieldData1[tI1].Contents[tI2].Name = READONLY_LABEL Then ReadOnly1 := FieldData1[tI1].Contents[tI2].Value;
    End;

    If FieldKind1 = 'data' Then FieldKind2 := fkData
    Else If FieldKind1 = 'calculated' Then FieldKind2 := fkCalculated
    Else If FieldKind1 = 'lookup' Then FieldKind2 := fkLookup
    Else If FieldKind1 = 'internalcalc' Then FieldKind2 := fkInternalCalc
    Else If FieldKind1 = 'aggregate' Then FieldKind2 := fkAggregate
    Else FieldKind2 := fkData;

    DisplayLabel2 := DisplayLabel1;
    EditMask2 := EditMask1;
    If DisplayWidth1 = '' Then DisplayWidth2 := 0 Else DisplayWidth2 := StrToInt(DisplayWidth1);
    If ReadOnly1 = TRUE_LABEL Then ReadOnly2 := True Else ReadOnly2 := False;
    Field1 := FDataSet.FieldByName(FieldName1);
    Field1.FieldKind := FieldKind2;
    Field1.DisplayWidth := DisplayWidth2;
    Field1.DisplayLabel := DisplayLabel2;
    Field1.EditMask := EditMask2;
    Field1.ReadOnly := ReadOnly2;
  End;
End;

Procedure TXMLDataset.DoIndexDefinitions(Var ABeginPos: Integer; Const AEndPos: Integer); //I have not tested this routine
{$IFDEF TTABLE_DERIVATIVE}
Var
  tI1: Integer;
  tI2: Integer;
  BeginPos1: Integer;
  EndPos1: Integer;
  IndexData1: Array[0..255] Of TLocalFieldList;
  IndexName1: String;
  IndexFields1: String;
  DisplayLabel1: String;
  IndexOptions1: TIndexOptions
{$ENDIF}
Begin
{$IFDEF TTABLE_DERIVATIVE}
  //TDataSet does not have Indexes, only TTable and its derivatives do.
  //So, unless we are using a TTable (or derivative), even if there were indexes, we throw them out
  IndexCount1 := -1;
  BeginPos1 := ABeginPos;
  While (BeginPos1 < AEndPos) Do Begin
    If GetBounds(INDEXDEF_BOUNDS, ABeginPos, AEndPos, BeginPos1, EndPos1) Then Begin
      Inc(IndexCount1);
      DoAtribAndValue(BeginPos1, EndPos1, IndexData1[IndexCount1].AttribCount, IndexData1[IndexCount1].Contents);
    End Else Break;
    ABeginPos := BeginPos1;
  End;

  For tI1 := 0 To IndexCount1 Do Begin
    IndexName1 := '';
    IndexFields1 := '';
    DisplayLabel1 := '';
    IndexOptions1 := [];
    For tI2 := 0 To IndexData1[tI1].AttribCount - 1 Do Begin
      If IndexData1[tI1].Contents[tI2].Name = NAME_LABEL Then IndexName1 := IndexData1[tI1].Contents[tI2].Value;
      If IndexData1[tI1].Contents[tI2].Name = INDEXFIELDS_LABEL Then IndexFields1 := IndexData1[tI1].Contents[tI2].Value;
      If IndexData1[tI1].Contents[tI2].Name = DISPLAYNAME_LABEL Then DisplayLabel1 := IndexData1[tI1].Contents[tI2].Value;
      If IndexData1[tI1].Contents[tI2].Name = SORTORDER_LABEL Then Begin
        If IndexData1[tI1].Contents[tI2].Value = SORTORDER_LABEL Then Include(IndexOptions1, [ixDescending]) Else Exclude(IndexOptions1, [ixDescending]);
      End;
      If IndexData1[tI1].Contents[tI2].Name = CASESENSITIVE_LABEL Then Begin
        If IndexData1[tI1].Contents[tI2].Value = TRUE_LABEL Then Include(IndexOptions1, [ixCaseInsensitive]) Else Exclude(IndexOptions1, [ixCaseInsensitive]);
      End;
      If IndexData1[tI1].Contents[tI2].Name = NONMAINTAINED_LABEL Then Begin
        If IndexData1[tI1].Contents[tI2].Value = TRUE_LABEL Then Include(IndexOptions1, [ixNonMaintained]) Else Exclude(IndexOptions1, [ixNonMaintained]);
      End;
      If IndexData1[tI1].Contents[tI2].Name = UNIQUE_LABEL Then Begin
        If IndexData1[tI1].Contents[tI2].Value = TRUE_LABEL Then Include(IndexOptions1, [ixUnique]) Else Exclude(IndexOptions1, [ixUnique]);
      End;
    End;
    FDataSet.IndexDefs.Add(IndexName1, IndexFields1, IndexOptions1);
    FDataSet.IndexDefs[tI1].DisplayLabel := DisplayLabel1;
  End;

{$ENDIF}
End;

Procedure TXMLDataset.DoRecordData(Var ABeginPos: Integer; Const AEndPos: Integer);
Var
  BeginPos1: Integer;
  EndPos1: Integer;
  RecordCount1: Integer;
Begin
  FDataSet.Open;
  RecordCount1 := 0;
  While (ABeginPos < AEndPos) And GetBounds(ROWDATA_BOUNDS, ABeginPos, AEndPos, BeginPos1, EndPos1) Do Begin
    DoFieldData(BeginPos1, EndPos1);
    ABeginPos := BeginPos1;
    Inc(RecordCount1);
  End;
  FDataSet.First;
  If (FRecordCount > 0) And (FRecordCount <> RecordCount1) Then Raise EXMLDatasetError.Create('Insufficient Record Count Received');
End;

Procedure TXMLDataset.DoFieldData(Var ABeginPos: Integer; Const AEndPos: Integer);
  Procedure DoCData(ABeginPos: Integer; AEndPos: Integer; Var AString: String);
  Var
    BeginPos1: Integer;
    EndPos1: Integer;
    String1: String;
  Begin
    // RJ AString := '';
    If GetBounds(CDATA_BOUNDS, ABeginPos, AEndPos, BeginPos1, EndPos1) Then Begin
      CopyToString(BeginPos1 + CDATA_BOUNDS[0].Length - 1, EndPos1, String1);
      // RJ FBase64MIME.Decode (string1, AString);
      AString := MimeDecodeString(String1);
    End;
  End;
Var
  Count1: Integer;
  FieldName1: String;
  Size1: Integer;
  BeginPos1: Integer;
  EndPos1: Integer;
  String1: String;
  Contents1: Array[0..255] Of TParamAndValue;
Begin
  FDataSet.Append;
  While (ABeginPos < AEndPos) And GetBounds(FIELD_BOUNDS, ABeginPos, AEndPos, BeginPos1, EndPos1) Do Begin
    Count1 := 0;
    DoAtribAndValue(BeginPos1, EndPos1, Count1, Contents1);
    If Count1 = 2 Then Begin
      If (Contents1[0].Name = NAME_LABEL) And (Contents1[1].Name = VALUE_LABEL) Then Begin
        FieldName1 := Contents1[0].Value;
        String1 := Contents1[1].Value;
        FDataSet.FieldByName(FieldName1).AsString := String1;
      End Else If (Contents1[0].Name = NAME_LABEL) And (Contents1[1].Name = SIZE_LABEL) Then Begin
        FieldName1 := Contents1[0].Value;
        Size1 := StrToInt(Contents1[1].Value);
        If Size1 < 0 Then FDataSet.FieldByName(FieldName1).AsVariant := Null
        Else If Size1 = 0 Then FDataSet.FieldByName(FieldName1).AsString := ''
        Else Begin
          FindFirst(FIELD_END_FULL, BeginPos1, AEndPos, EndPos1); //marker begin position.
          EndPos1 := EndPos1 + LEN_FIELD_END_FULL - 1;

          DoCData(BeginPos1, EndPos1, String1);
          If Length(String1) <> Size1 Then Raise EXMLDatasetError.Create('Unexpected CDATA Size');
          FDataSet.FieldByName(FieldName1).AsString := String1;
        End;
      End Else Raise EXMLDatasetError.Create('Unexpected parameter(s)');
    End Else Raise EXMLDatasetError.Create('Insufficient number of parameters');
    ABeginPos := EndPos1;
  End;
  FDataSet.Post;
End;

Function TXMLDataset.GetBounds(Const ABounds: Array Of TBoundsInfo; Const ABeginPos, AEndPos: Integer; Var ABoundsBegin, ABoundsEnd: Integer): Boolean;
Begin
  Result := False;
  FindFirst(ABounds[0].Name, ABeginPos, AEndPos, ABoundsBegin);
  If (ABoundsBegin < ABeginPos) Then Exit;
  FindFirst(ABounds[1].Name, ABoundsBegin, AEndPos, ABoundsEnd);
  If (ABoundsEnd > AEndPos) Or (ABoundsEnd <= ABoundsBegin) Then Exit;
  Result := True;
End;

Procedure TXMLDataset.GetMetaDataProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
Var
  tI1: Integer;
  EndPos1: Integer;
  AttribCount1: Integer;
  HasFieldDefs1: Boolean;
  HasIndexDefs1: Boolean;
  HasRecordData1: Boolean;
  HasChanges1: Boolean;
  Array1: Array[0..255] Of TParamAndValue;
Begin
  AttribCount1 := 0;
  HasFieldDefs1 := False;
  HasIndexDefs1 := False;
  HasRecordData1 := False;
  HasChanges1 := False;

  FindFirst(TAG_END, ABeginPos, AEndPos, EndPos1); //marker begin position.
  DoAtribAndValue(ABeginPos, EndPos1, AttribCount1, Array1);
  For tI1 := 0 To AttribCount1 - 1 Do Begin
    If Array1[tI1].Name = FIELDDEFS_LABEL Then HasFieldDefs1 := (Array1[tI1].Value = YES_LABEL)
    Else If Array1[tI1].Name = INDEXDEFS_LABEL Then HasIndexDefs1 := (Array1[tI1].Value = YES_LABEL)
    Else If Array1[tI1].Name = RECORDDATA_LABEL Then HasRecordData1 := (Array1[tI1].Value = YES_LABEL)
    Else If Array1[tI1].Name = CHANGES_LABEL Then HasChanges1 := (Array1[tI1].Value = YES_LABEL);
  End;
  FHasFieldDefs := HasFieldDefs1;
  FHasIndexDefs := HasIndexDefs1;
  FHasRecordData := HasRecordData1;
  FHasChanges := HasChanges1;
End;

Procedure TXMLDataset.GetFieldDefinitionProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
Var
  tI1: Integer;
  EndPos1: Integer;
  AttribCount1: Integer;
  Array1: Array[0..255] Of TParamAndValue;
Begin
  FindFirst(TAG_END, ABeginPos, AEndPos, EndPos1); //marker begin position.
  DoAtribAndValue(ABeginPos, EndPos1, AttribCount1, Array1);

  For tI1 := 0 To AttribCount1 - 1 Do Begin
    If Array1[tI1].Name = COUNT_LABEL Then FFieldDefCount := StrToInt(Array1[tI1].Value);
  End;
End;

Procedure TXMLDataset.GetIndexDefinitionProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
Var
  tI1: Integer;
  EndPos1: Integer;
  AttribCount1: Integer;
  Array1: Array[0..255] Of TParamAndValue;
Begin
  FindFirst(TAG_END, ABeginPos, AEndPos, EndPos1); //marker begin position.
  DoAtribAndValue(ABeginPos, EndPos1, AttribCount1, Array1);

  For tI1 := 0 To AttribCount1 - 1 Do Begin
    If Array1[tI1].Name = COUNT_LABEL Then FIndexDefCount := StrToInt(Array1[tI1].Value);
  End;
End;

Procedure TXMLDataset.GetRecordDataProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
Var
  tI1: Integer;
  EndPos1: Integer;
  AttribCount1: Integer;
  Array1: Array[0..255] Of TParamAndValue;
Begin
  FindFirst(TAG_END, ABeginPos, AEndPos, EndPos1); //marker begin position.
  DoAtribAndValue(ABeginPos, EndPos1, AttribCount1, Array1);

  For tI1 := 0 To AttribCount1 - 1 Do Begin
    If Array1[tI1].Name = COUNT_LABEL Then FRecordCount := StrToInt(Array1[tI1].Value);
  End;
End;

Procedure TXMLDataset.GetDataPacketProperties(Var ABeginPos: Integer; Const AEndPos: Integer);
Var
  tI1: Integer;
  Day1: Integer;
  Month1: Integer;
  Year1: Integer;
  Hour1: Integer;
  Min1: Integer;
  Sec1: Integer;
  MSec1: Integer;
  EndPos1: Integer;
  AttribCount1: Integer;
  Version1: Single;
  TableName1: String;
  Time1: TDateTime;
  Date1: TDateTime;
  Array1: Array[0..255] Of TParamAndValue;
Begin
  Day1 := 0;
  Month1 := 0;
  Year1 := 0;
  Hour1 := 0;
  Min1 := 0;
  Sec1 := 0;
  MSec1 := 0;
  EndPos1 := 0;
  Version1 := FFileVersion;
  FindFirst(TAG_END, ABeginPos, AEndPos, EndPos1); //marker begin position.
  DoAtribAndValue(ABeginPos, EndPos1, AttribCount1, Array1);
  For tI1 := 0 To AttribCount1 - 1 Do Begin
    If Array1[tI1].Name = VERSION_LABEL Then Version1 := StrToFloat(Array1[tI1].Value)
    Else If Array1[tI1].Name = DESTINATION_LABEL Then TableName1 := Array1[tI1].Value
    Else If Array1[tI1].Name = DAY_LABEL Then Day1 := StrToInt(Array1[tI1].Value)
    Else If Array1[tI1].Name = MONTH_LABEL Then Month1 := StrToInt(Array1[tI1].Value)
    Else If Array1[tI1].Name = YEAR_LABEL Then Year1 := StrToInt(Array1[tI1].Value)
    Else If Array1[tI1].Name = HOUR_LABEL Then Hour1 := StrToInt(Array1[tI1].Value)
    Else If Array1[tI1].Name = MIN_LABEL Then Min1 := StrToInt(Array1[tI1].Value)
    Else If Array1[tI1].Name = SEC_LABEL Then Sec1 := StrToInt(Array1[tI1].Value)
    Else If Array1[tI1].Name = MSEC_LABEL Then MSec1 := StrToInt(Array1[tI1].Value);
  End;

  Time1 := EncodeTime(Hour1, Min1, Sec1, MSec1);
  Date1 := EncodeDate(Year1, Month1, Day1);
  FPacketTime := Date1 + Time1;
  FPacketVersion := Version1;
End;

Function TXMLDataset.Execute: Boolean;
Var
  Fields1: Integer;
  Fields2: Integer;
  Indexes1: Integer;
  Indexes2: Integer;
  MetaData1: Integer;
  MetaData2: Integer;
  RecordData1: Integer;
  RecordData2: Integer;
  DataPacket1: Integer;
  DataPacket2: Integer;
  ShortDateFormat1: String;
  ShortTimeFormat1: String;

  CurrencyString1: String;
  CurrencyFormat1: Byte;
  NegCurrFormat1: Byte;
  ThousandSeparator1: Char;
  DecimalSeparator1: Char;
  CurrencyDecimals1: Byte;
  DateSeparator1: Char;
  TimeSeparator1: Char;
  TimeAMString1: String;
  TimePMString1: String;
  LongTimeFormat1: String;
  TwoDigitYearCenturyWindow1: Word;
  ListSeparator1: Char;

  ControlsDisabled1: Boolean;
Begin
  Result := True;
  CurrencyDecimals1 := CurrencyDecimals;
  CurrencyFormat1 := CurrencyFormat;
  CurrencyString1 := CurrencyString;
  DateSeparator1 := DateSeparator;
  DecimalSeparator1 := DecimalSeparator;
  ListSeparator1 := ListSeparator;
  LongTimeFormat1 := LongTimeFormat;
  NegCurrFormat1 := NegCurrFormat;
  ShortDateFormat1 := ShortDateFormat;
  ShortTimeFormat1 := ShortTimeFormat;
  ThousandSeparator1 := ThousandSeparator;
  TimeAMString1 := TimeAMString;
  TimePMString1 := TimePMString;
  TimeSeparator1 := TimeSeparator;
  TwoDigitYearCenturyWindow1 := TwoDigitYearCenturyWindow;

  CurrencyDecimals := 2;
  CurrencyFormat := 0;
  CurrencyString := '$';
  DateSeparator := '-';
  DecimalSeparator := '.';
  ListSeparator := ';';
  LongTimeFormat := LongTimeFormat;
  NegCurrFormat := 1;
  ShortDateFormat := 'yyyy-mm-dd';
  ShortTimeFormat := 'hh:nn:ss:zzz';
  ThousandSeparator := ',';
  TimeAMString := 'AM';
  TimePMString := 'PM';
  TimeSeparator := ':';
  TwoDigitYearCenturyWindow := 50;

  Try
    If Not Assigned(FDataSet) Then Raise EXMLDatasetError.Create('Dataset not assigned');
    ControlsDisabled1 := FDataSet.ControlsDisabled;
    If Not ControlsDisabled1 Then FDataSet.DisableControls;
    Try
      If GetBounds(DATAPACKET_BOUNDS, 0, MaxInt, DataPacket1, DataPacket2) Then Begin
        GetDataPacketProperties(DataPacket1, DataPacket2);
        If FPacketVersion > FFileVersion Then Raise EXMLDatasetError.Create('Incompatible file verion');
        If GetBounds(METADATA_BOUNDS, DataPacket1, DataPacket2, MetaData1, MetaData2) Then Begin
          GetMetaDataProperties(MetaData1, MetaData2);
          If GetBounds(FIELDDEFS_BOUNDS, MetaData1, MetaData2, Fields1, Fields2) Then Begin
            GetFieldDefinitionProperties(Fields1, Fields2);
            DoFieldDefinitions(Fields1, Fields2); //Process Field Defintions
            If GetBounds(INDEXDEFS_BOUNDS, Fields2 + FIELDDEFS_BOUNDS[1].Length - 1, MetaData2 - METADATA_BOUNDS[1].Length, Indexes1, Indexes2) Then Begin
              GetIndexDefinitionProperties(Indexes1, Indexes2);
              DoIndexDefinitions(Indexes1, Indexes2);
              If GetBounds(RECORDDATA_BOUNDS, Indexes2 + INDEXDEFS_BOUNDS[1].Length - 1, DataPacket2, RecordData1, RecordData2) Then Begin
                GetRecordDataProperties(RecordData1, RecordData2);
                DoRecordData(RecordData1, RecordData2);
              End;
            End Else If GetBounds(RECORDDATA_BOUNDS, Fields2 + FIELDDEFS_BOUNDS[1].Length - 1, DataPacket2, RecordData1, RecordData2) Then Begin
              GetRecordDataProperties(RecordData1, RecordData2);
              DoRecordData(RecordData1, RecordData2);
            End;
          End;
        End;
      End;
    Except
      On e: Exception Do Result := False;
    End;
    If Not ControlsDisabled1 Then FDataSet.EnableControls;
  Finally
    CurrencyDecimals := CurrencyDecimals1;
    CurrencyFormat := CurrencyFormat1;
    CurrencyString := CurrencyString1;
    DateSeparator := DateSeparator1;
    DecimalSeparator := DecimalSeparator1;
    ListSeparator := ListSeparator1;
    LongTimeFormat := LongTimeFormat1;
    NegCurrFormat := NegCurrFormat1;
    ShortDateFormat := ShortDateFormat1;
    ShortTimeFormat := ShortTimeFormat1;
    ThousandSeparator := ThousandSeparator1;
    TimeAMString := TimeAMString1;
    TimePMString := TimePMString1;
    TimeSeparator := TimeSeparator1;
    TwoDigitYearCenturyWindow := TwoDigitYearCenturyWindow1;
  End;
End;

End.

