Unit XMLDatasetUtils;

Interface

Uses
  Windows,
  SysUtils,
  Classes,
  Db;

Type
  TBoundsInfo = Record
    Name: String;
    Length: Integer;
  End;

  TParamAndValue = Record
    Name: String;
    Value: String;
  End;

  ESeekInXMLFileError = Exception;

  TSeekInXMLFile = Class(TComponent)
  Private
    FBeginPos: Integer;
    FEndPos: Integer;
    FText: PChar;
    FFileName: TFileName;
    FTextSize: Integer;
    FInMemory: Boolean;
    FFound: Boolean;
    FIsCaseSensitive: Boolean;
    FJump: Integer;
    FJumpTable: Array[#0..#255] Of Integer;
    FSubstring: String;
    FSubstringLenth: Integer;
    Function GetText: String;
    Procedure SetText(AString: String);
  Public
    Constructor Create(AOwner: TComponent); Override;
    Destructor Destroy; Override;
    Procedure Clear;
    Procedure SaveToStream(AStream: TStream);
    Procedure FindFirst(Const ASubstring: String; Const ABeginPos, AEndPos: Integer; Var AResult: Integer);
    Procedure CopyToString(Const ABeginPos, AEndPos: Integer; Var AResult: String); //DOES include ABeginPos and AEndPos
    Procedure FindNext(Var AResult: Integer);
    Procedure LoadFromFile(AFileName: TFileName);
    Property Text: String Read GetText Write SetText;
  Published
    Property CaseSensitive: Boolean Read FIsCaseSensitive Write FIsCaseSensitive;
    Property FileName: TFileName Read FFileName Write LoadFromFile;
  End;

Const
  FILE_VERSION: Single = 2.000;
  QUOTE_CHAR = '"';
  EXCLAIM_MARK = '!';
  LF_CHAR = #10;
  CR_CHAR = #13;
  CRLF_CHARS = CR_CHAR + LF_CHAR;
  EQUAL_CHAR: Char = '=';
  SPACE_CHAR = #32;
  TAB_CHAR = #9;
  FORWARD_SLASH = '/';
  QUESTION_MARK = '?';
  OPEN_BRACKET = '[';
  CLOSE_BRACKET = ']';
  TAG_BEGIN = '<';
  TAG_END = '>';

  CASESENSITIVE_LABEL = 'casesensitive';
  ASCENDING_LABEL = 'ascending';
  CDATA_LABEL = 'CDATA';
  CHANGE_LABEL = 'change'; //this is used for 'changes'
  CHANGES_LABEL = 'changes'; //this is used for 'changes'
  COUNT_LABEL = 'count';
  CRITERIA_LABEL = 'criteria'; //this is used for 'changes'
  DATAPACKET_LABEL = 'datapacket';
  DATATYPE_LABEL = 'datatype';
  DAY_LABEL = 'day';
  DESCENDING_LABEL = 'descending';
  DESTINATION_LABEL = 'destination'; //this stores the table name
  DISPLAYLABEL_LABEL = 'displaylabel';
  DISPLAYNAME_LABEL = 'displayname';
  DISPLAYWIDTH_LABEL = 'displaywidth';
  EDITMASK_LABEL = 'editmask';
  FALSE_LABEL = 'false';
  FIELD_LABEL = 'field';
  FIELDDEF_LABEL = 'fielddef';
  FIELDDEFS_LABEL = 'fielddefs';
  FIELDINDEX_LABEL = 'fieldindex';
  FIELDKIND_LABEL = 'fieldkind';
  FIELDSIZE_LABEL = 'fieldsize';
  HOUR_LABEL = 'hour';
  INDEX_LABEL = 'index';
  INDEXDEF_LABEL = 'indexdef';
  INDEXDEFS_LABEL = 'indexdefs';
  INDEXFIELDS_LABEL = 'indexfields';
  INDEXFIELDNAMES_LABEL = 'indexfieldnames';
  INDEXNAME_LABEL = 'indexname';
  INDEXNO_LABEL = 'indexno';
  NONMAINTAINED_LABEL = 'nonmaintained';
  METADATA_LABEL = 'metadata';
  MIN_LABEL = 'min';
  MONTH_LABEL = 'month';
  MSEC_LABEL = 'msec';
  NAME_LABEL = 'name';
  NEWVALUES_LABEL = 'newvalues'; //this is used for 'changes'
  NO_LABEL = 'no';
  PRIMARYKEY_LABEL = 'primarykey';
  READONLY_LABEL = 'readonly';
  RECORDDATA_LABEL = 'recorddata';
  REQUIRED_LABEL = 'required';
  ROW_LABEL = 'row';
  SEC_LABEL = 'sec';
  SIZE_LABEL = 'size';
  SORTORDER_LABEL = 'sortorder';
  STANDALONE_LABEL = 'standalone';
  TRUE_LABEL = 'true';
  TYPE_LABEL = 'type';
  UNIQUE_LABEL = 'unique';
  UPDATE_LABEL = 'update'; //this is used for 'changes'
  INSERT_LABEL = 'insert'; //this is used for 'changes'
  DELETE_LABEL = 'delete'; //this is used for 'changes'
  VALUE_LABEL = 'value';
  VERSION_LABEL = 'version';
  XML_LABEL = 'xml';
  YEAR_LABEL = 'year';
  YES_LABEL = 'yes';

  CHANGES_BEGIN = TAG_BEGIN + CHANGES_LABEL;
  CHANGES_END = TAG_BEGIN + CHANGES_LABEL + TAG_END;

  CHANGE_BEGIN = TAG_BEGIN + CHANGE_LABEL;
  CHANGE_END = TAG_BEGIN + CHANGE_LABEL + TAG_END;

  CRITERIA_BEGIN = TAG_BEGIN + CRITERIA_LABEL;
  CRITERIA_END = TAG_BEGIN + CRITERIA_LABEL + TAG_END;

  CDATA_BEGIN = TAG_BEGIN + EXCLAIM_MARK + OPEN_BRACKET + CDATA_LABEL + OPEN_BRACKET; //<![cdata[
  CDATA_END = CLOSE_BRACKET + CLOSE_BRACKET + TAG_END; //]]>

  DATAPACKET_BEGIN = TAG_BEGIN + DATAPACKET_LABEL;
  DATAPACKET_END = TAG_BEGIN + FORWARD_SLASH + DATAPACKET_LABEL + TAG_END;

  EXTRATAG_END = FORWARD_SLASH + TAG_END;

  FIELDDEFS_BEGIN = TAG_BEGIN + FIELDDEFS_LABEL;
  FIELDDEFS_END = TAG_BEGIN + FORWARD_SLASH + FIELDDEFS_LABEL + TAG_END;

  FIELDDEF_BEGIN = TAG_BEGIN + FIELDDEF_LABEL;
  FIELDDEF_END = FORWARD_SLASH + TAG_END;

  FIELD_BEGIN = TAG_BEGIN + FIELD_LABEL;
  FIELD_END = TAG_END;
  FIELD_END_FULL = TAG_BEGIN + FORWARD_SLASH + FIELD_LABEL + TAG_END;

  PRIMARYKEY_BEGIN = TAG_BEGIN + PRIMARYKEY_LABEL;
  PRIMARYKEY_END = TAG_BEGIN + FORWARD_SLASH + PRIMARYKEY_LABEL + TAG_END;

  INDEX_BEGIN = TAG_BEGIN + INDEX_LABEL;
  INDEX_END = TAG_END;
  INDEX_END_FULL = TAG_BEGIN + FORWARD_SLASH + INDEX_LABEL + TAG_END;

  INDEXDEFS_BEGIN = TAG_BEGIN + INDEXDEFS_LABEL;
  INDEXDEFS_END = TAG_BEGIN + FORWARD_SLASH + INDEXDEFS_LABEL + TAG_END;

  INDEXDEF_BEGIN = TAG_BEGIN + INDEXDEF_LABEL;
  INDEXDEF_END = FORWARD_SLASH + TAG_END;

  METADATA_BEGIN = TAG_BEGIN + METADATA_LABEL;
  METADATA_END = TAG_BEGIN + FORWARD_SLASH + METADATA_LABEL + TAG_END;

  RECORDDATA_BEGIN = TAG_BEGIN + RECORDDATA_LABEL;
  RECORDDATA_END = TAG_BEGIN + FORWARD_SLASH + RECORDDATA_LABEL + TAG_END;

  ROW_BEGIN = TAG_BEGIN + ROW_LABEL + TAG_END;
  ROW_END = TAG_BEGIN + FORWARD_SLASH + ROW_LABEL + TAG_END;

  XML_BEGIN = TAG_BEGIN + QUESTION_MARK + XML_LABEL;
  XML_END = QUESTION_MARK + TAG_END;

  LEN_CDATA_BEGIN = Length(CDATA_BEGIN);
  LEN_CDATA_END = Length(CDATA_END);
  LEN_DATAPACKET_BEGIN = Length(DATAPACKET_BEGIN);
  LEN_DATAPACKET_END = Length(DATAPACKET_END);
  LEN_DOUBLEQUOTE = Length(QUOTE_CHAR);
  LEN_EXTRATAG_END = Length(EXTRATAG_END);
  LEN_FIELDDEFS_BEGIN = Length(FIELDDEFS_BEGIN);
  LEN_FIELDDEFS_END = Length(FIELDDEFS_END);
  LEN_FIELDDEF_BEGIN = Length(FIELDDEF_BEGIN);
  LEN_FIELDDEF_END = Length(FIELDDEF_END);
  LEN_FIELD_BEGIN = Length(FIELD_BEGIN);
  LEN_FIELD_END = Length(FIELD_END);
  LEN_FIELD_END_FULL = Length(FIELD_END_FULL);
  LEN_INDEXDEFS_BEGIN = Length(INDEXDEFS_BEGIN);
  LEN_INDEXDEFS_END = Length(INDEXDEFS_END);
  LEN_INDEXDEF_BEGIN = Length(INDEXDEF_BEGIN);
  LEN_INDEXDEF_END = Length(INDEXDEF_END);

  LEN_INDEX_BEGIN = Length(INDEX_BEGIN);
  LEN_INDEX_END = Length(INDEX_END);

  LEN_METADATA_BEGIN = Length(METADATA_BEGIN);
  LEN_METADATA_END = Length(METADATA_END);
  LEN_RECORDDATA_BEGIN = Length(RECORDDATA_BEGIN);
  LEN_RECORDDATA_END = Length(RECORDDATA_END);
  LEN_ROW_BEGIN = Length(ROW_BEGIN);
  LEN_ROW_END = Length(ROW_END);
  LEN_TAG_BEGIN = Length(TAG_BEGIN);
  LEN_TAG_END = Length(TAG_END);
  LEN_CHANGES_BEGIN = Length(CHANGES_BEGIN);
  LEN_CHANGES_END = Length(CHANGES_END);
  LEN_CHANGE_BEGIN = Length(CHANGE_BEGIN);
  LEN_CHANGE_END = Length(CHANGE_END);
  LEN_CRITERIA_BEGIN = Length(CRITERIA_BEGIN);
  LEN_CRITERIA_END = Length(CRITERIA_END);

  CRITERIA_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: CRITERIA_BEGIN; Length: LEN_CRITERIA_BEGIN), (Name: CRITERIA_END; Length: LEN_CRITERIA_END));
  CHANGES_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: CHANGES_BEGIN; Length: LEN_CHANGES_BEGIN), (Name: CHANGES_END; Length: LEN_CHANGES_END));
  CHANGE_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: CHANGE_BEGIN; Length: LEN_CHANGE_BEGIN), (Name: CHANGE_END; Length: LEN_CHANGE_END));

  CDATA_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: CDATA_BEGIN; Length: LEN_CDATA_BEGIN), (Name: CDATA_END; Length: LEN_CDATA_END));
  DATAPACKET_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: DATAPACKET_BEGIN; Length: LEN_DATAPACKET_BEGIN), (Name: DATAPACKET_END; Length: LEN_DATAPACKET_END));
  EXTRATAGDATA_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: TAG_BEGIN; Length: LEN_TAG_BEGIN), (Name: EXTRATAG_END; Length: LEN_EXTRATAG_END));
  FIELDDEFS_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: FIELDDEFS_BEGIN; Length: LEN_FIELDDEFS_BEGIN), (Name: FIELDDEFS_END; Length: LEN_FIELDDEFS_END));
  FIELD_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: FIELD_BEGIN; Length: LEN_FIELD_BEGIN), (Name: FIELD_END; Length: LEN_FIELD_END));
  FIELDDEF_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: FIELDDEF_BEGIN; Length: LEN_FIELDDEF_BEGIN), (Name: FIELDDEF_END; Length: LEN_FIELDDEF_END));
  INDEX_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: INDEX_BEGIN; Length: LEN_INDEX_BEGIN), (Name: INDEX_END; Length: LEN_INDEX_END));
  INDEXDEFS_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: INDEXDEFS_BEGIN; Length: LEN_INDEXDEFS_BEGIN), (Name: INDEXDEFS_END; Length: LEN_INDEXDEFS_END));
  INDEXDEF_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: INDEXDEF_BEGIN; Length: LEN_INDEXDEF_BEGIN), (Name: INDEXDEF_END; Length: LEN_INDEXDEF_END));
  METADATA_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: METADATA_BEGIN; Length: LEN_METADATA_BEGIN), (Name: METADATA_END; Length: LEN_METADATA_END));
  RECORDDATA_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: RECORDDATA_BEGIN; Length: LEN_RECORDDATA_BEGIN), (Name: RECORDDATA_END; Length: LEN_RECORDDATA_END));
  ROWDATA_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: ROW_BEGIN; Length: LEN_ROW_BEGIN), (Name: ROW_END; Length: LEN_ROW_END));
  TAGDATA_BOUNDS: Array[0..1] Of TBoundsInfo = ((Name: TAG_BEGIN; Length: LEN_TAG_BEGIN), (Name: TAG_END; Length: LEN_TAG_END));

Function DataTypeToString(AField: TField): String;
Function StringToDataType(AType: String): TFieldType;

Implementation

Function DataTypeToString(AField: TField): String;
Begin
  Case AField.DataType Of
    ftADT: Result := 'adt';
    ftArray: Result := 'array';
    ftAutoInc: Result := 'autoinc';
    ftBCD: Result := 'bcd';
    ftBlob: Result := 'blob';
    ftBoolean: Result := 'boolean';
    ftBytes: Result := 'bytes';
    ftCurrency: Result := 'currency';
    ftCursor: Result := 'cursor';
    ftDataSet: Result := 'dataset';
    ftDate: Result := 'date';
    ftDateTime: Result := 'datetime';
    ftDBaseOle: Result := 'dbaseole';
    ftFixedChar: Result := 'fixedchar';
    ftFloat: Result := 'float';
    ftFmtMemo: Result := 'fmtmemo';
    ftGraphic: Result := 'graphic';
    {$IFDEF ver130}
    ftGuid: Result := 'guid';
    ftIDispatch: Result := 'idispatch';
    {$ENDIF}
    ftInteger: Result := 'integer';
    {$IFDEF ver130}
    ftInterface: Result := 'interface';
    {$ENDIF}
    ftLargeInt: Result := 'largeint';
    ftMemo: Result := 'memo';
    {$IFDEF ver130}
    ftOraBlob: Result := 'orablob';
    ftOraClob: Result := 'oraclob';
    {$ENDIF}
    ftParadoxOle: Result := 'paradoxole';
    ftReference: Result := 'reference';
    ftSmallInt: Result := 'smallint';
    ftString: Result := 'string';
    ftTime: Result := 'time';
    ftTypedBinary: Result := 'typedbinary';
    ftUnknown: Result := 'unknown';
    ftVarBytes: Result := 'varbytes';
    {$IFDEF ver130}
    ftVariant: Result := 'variant';
    {$ENDIF}
    //FtWideString: Result := 'widestring';
    ftWideString: Result := 'string';
    ftWord: Result := 'word';
    Else
      Result := 'unknown';
  End;
End;

Function StringToDataType(AType: String): TFieldType;
Begin
  If AType = 'adt' Then Result := ftADT
  Else If AType = 'array' Then Result := ftArray
  Else If AType = 'autoinc' Then Result := ftAutoInc
  Else If AType = 'bcd' Then Result := ftBCD
  Else If AType = 'blob' Then Result := ftBlob
  Else If AType = 'boolean' Then Result := ftBoolean
  Else If AType = 'bytes' Then Result := ftBytes
  Else If AType = 'currency' Then Result := ftCurrency
  Else If AType = 'cursor' Then Result := ftCursor
  Else If AType = 'dataset' Then Result := ftDataSet
  Else If AType = 'date' Then Result := ftDate
  Else If AType = 'datetime' Then Result := ftDateTime
  Else If AType = 'dbaseole' Then Result := ftDBaseOle
  Else If AType = 'fixedchar' Then Result := ftFixedChar
  Else If AType = 'float' Then Result := ftFloat
  Else If AType = 'fmtmemo' Then Result := ftFmtMemo
  Else If AType = 'graphic' Then Result := ftGraphic
    {$IFDEF ver130}
  Else If AType = 'guid' Then Result := ftGuid
  Else If AType = 'idispatch' Then Result := ftIDispatch
    {$ENDIF}
  Else If AType = 'integer' Then Result := ftInteger
    {$IFDEF ver130}
  Else If AType = 'interface' Then Result := ftInterface
    {$ENDIF}
  Else If AType = 'largeint' Then Result := ftLargeInt
  Else If AType = 'memo' Then Result := ftMemo
    {$IFDEF ver130}
  Else If AType = 'orablob' Then Result := ftOraBlob
  Else If AType = 'oraclob' Then Result := ftOraClob
    {$ENDIF}
  Else If AType = 'paradoxole' Then Result := ftParadoxOle
  Else If AType = 'reference' Then Result := ftReference
  Else If AType = 'smallint' Then Result := ftSmallInt
  Else If AType = 'string' Then Result := ftString
  Else If AType = 'time' Then Result := ftTime
  Else If AType = 'typedbinary' Then Result := ftTypedBinary
  Else If AType = 'unknown' Then Result := ftUnknown
  Else If AType = 'varbytes' Then Result := ftVarBytes
    {$IFDEF ver130}
  Else If AType = 'variant' Then Result := ftVariant
    {$ENDIF}
  Else If AType = 'widestring' Then Result := ftWideString
  Else If AType = 'word' Then Result := ftWord
  Else Result := ftUnknown;
End;

Var
  ComparisonTable: Array[#0..#255] Of Char;

Constructor TSeekInXMLFile.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  FText := Nil;
  FFileName := '';
  FTextSize := 0;
  FJump := 0;
  FIsCaseSensitive := False;
  FInMemory := False;
  FSubstringLenth := 0;
  FBeginPos := 0;
  FEndPos := 0;
  FSubstring := '';
End;

Destructor TSeekInXMLFile.Destroy;
Begin
  Clear;
  Inherited Destroy;
End;

Procedure TSeekInXMLFile.Clear;
Begin
  If (Not FInMemory) And (FText <> Nil) Then UnMapViewOfFile(FText);
  FText := Nil;
  FFileName := '';
  FTextSize := 0;
  FJump := 0;
  FIsCaseSensitive := False;
  FSubstringLenth := 0;
  FBeginPos := 0;
  FEndPos := 0;
  FSubstring := '';
End;

Function TSeekInXMLFile.GetText: String;
Begin
  Result := FText;
End;

Procedure TSeekInXMLFile.CopyToString(Const ABeginPos, AEndPos: Integer; Var AResult: String); //DOES include ABeginPos and AEndPos
Var
  tI1: Integer;
Begin
  AResult := '';
  For tI1 := ABeginPos To AEndPos Do AResult := AResult + FText[tI1];
End;

Procedure TSeekInXMLFile.SetText(AString: String);
Begin
  Clear;
  FInMemory := True;
  FText := PChar(AString);
  FTextSize := StrLen(FText);
End;

Procedure TSeekInXMLFile.LoadFromFile(AFileName: TFileName);
Var
  FileHandle1: Integer;
  FileMappingHandle1: THandle;
  HighSize1: Integer;
Begin
  Clear;
  FInMemory := False;
  If (AFileName = '') Then Raise ESeekInXMLFileError.Create('No FileName Given...');
  If Not FileExists(AFileName) Then Raise ESeekInXMLFileError.Create(AFileName + #13 + 'Does NOT Exist...');

  FileHandle1 := SysUtils.FileOpen(AFileName, fmOpenRead Or fmShareDenyNone); //Open File...
  If FileHandle1 = 0 Then Raise ESeekInXMLFileError.Create(AFileName + #13 + 'Could NOT be Opened For Reading...');

  FTextSize := GetFileSize(FileHandle1, @HighSize1); //Get File size;
  FileMappingHandle1 := CreateFileMapping(FileHandle1, Nil, page_readonly, 0, 0, Nil); //Get Map of File...

  If GetLastError = Error_Already_Exists Then Begin
    CloseHandle(FileMappingHandle1);
    FileMappingHandle1 := 0;
  End;

  If FileMappingHandle1 = 0 Then Begin
    CloseHandle(FileHandle1);
    Raise ESeekInXMLFileError.Create('MapFile Could not be opened.');
  End;

  FText := MapViewOfFile(FileMappingHandle1, File_Map_Read, 0, 0, 0);

  If FText <> Nil Then FFileName := AFileName;

  CloseHandle(FileMappingHandle1); //FileMappingHandle1 is no longer needed
  CloseHandle(FileHandle1); //FileHandle1 is no longer needed
End;

Procedure TSeekInXMLFile.FindFirst(Const ASubstring: String; Const ABeginPos, AEndPos: Integer; Var AResult: Integer);
Var
  tI1: Integer;
  tC1: Char;
Begin
  AResult := -1;
  FSubstring := ASubstring;
  FSubstringLenth := Length(FSubstring);
  FBeginPos := ABeginPos;
  FEndPos := AEndPos;

  If FText = Nil Then Raise ESeekInXMLFileError.Create('Nothing to search in...');
  If FSubstringLenth = 0 Then Raise ESeekInXMLFileError.Create('Nothing to search For...');

  If FBeginPos < 0 Then FBeginPos := 0;
  If FEndPos > FTextSize Then FEndPos := FTextSize;

  If FBeginPos >= FEndPos Then Raise ESeekInXMLFileError.Create('Begin position can not exceed End position');
  If FSubstringLenth > (FEndPos - FBeginPos) Then Exit;

  If FSubstringLenth = 0 Then Exit;
  For tC1 := #0 To #255 Do FJumpTable[tC1] := FSubstringLenth;

  Case FIsCaseSensitive Of
    True: Begin
        For tI1 := 1 To FSubstringLenth - 1 Do FJumpTable[FSubstring[tI1]] := FSubstringLenth - tI1;
        FJump := FJumpTable[FSubstring[FSubstringLenth]];
        FJumpTable[FSubstring[FSubstringLenth]] := 0;
      End;
    False: Begin
        For tI1 := 1 To FSubstringLenth - 1 Do FJumpTable[ComparisonTable[FSubstring[tI1]]] := FSubstringLenth - tI1;
        FJump := FJumpTable[ComparisonTable[FSubstring[FSubstringLenth]]];
        FJumpTable[ComparisonTable[FSubstring[FSubstringLenth]]] := 0;
      End;
  End;

  FFound := False;
  If FSubstringLenth <= (FEndPos - FBeginPos) Then FindNext(AResult);
End;

Procedure TSeekInXMLFile.FindNext(Var AResult: Integer);
Var
  tI1: Integer;
Begin
  FFound := False;
  AResult := -1;
  Inc(FBeginPos, FSubstringLenth - 1);
  Case FIsCaseSensitive Of
    True: Begin
        While FBeginPos < FEndPos Do Begin
          tI1 := FJumpTable[FText[FBeginPos]]; //test last character First
          If tI1 = 0 Then Begin //last char matches at least
            tI1 := FSubstringLenth - 1;
            While (tI1 > 0) And (FSubstring[FSubstringLenth - tI1] = FText[FBeginPos - tI1]) Do Dec(tI1);
            If (tI1 = 0) Then Begin
              FFound := True;
              AResult := FBeginPos - FSubstringLenth + 1;
              Inc(FBeginPos); //positioned now First character after match.
              Exit;
            End Else Inc(FBeginPos, FJump);
          End Else Inc(FBeginPos, tI1); //last char does not match
        End;
      End;
    False: Begin
        While FBeginPos < FEndPos Do Begin
          tI1 := FJumpTable[ComparisonTable[FText[FBeginPos]]];
          If tI1 = 0 Then Begin //last char matches at least
            tI1 := FSubstringLenth - 1;
            While (tI1 > 0) And (ComparisonTable[FSubstring[FSubstringLenth - tI1]] = ComparisonTable[FText[FBeginPos - tI1]]) Do Dec(tI1);
            If (tI1 = 0) Then Begin
              FFound := True;
              AResult := FBeginPos - FSubstringLenth + 1;
              Inc(FBeginPos); //positioned now First character after match.
              Exit;
            End Else Inc(FBeginPos, FJump);
          End Else Inc(FBeginPos, tI1); //last char does not match
        End;
      End;
  End;
End;

Procedure TSeekInXMLFile.SaveToStream(AStream: TStream);
Begin
  If FText <> Nil Then AStream.Write(FText^, FTextSize)
End;

Procedure MakeComparsonTable;
Var
  i: Byte;
Begin
  For i := 0 To 255 Do ComparisonTable[Char(i)] := AnsiLowerCase(Char(i))[1];
End;

Initialization
  MakeComparsonTable;
Finalization

End.

